@@ -39,8 +39,10 @@
 #include "arch-mips64n32.h"
 #include "arch-aarch64.h"
 #include "arch-parisc.h"
+#include "arch-parisc64.h"
 #include "arch-ppc.h"
 #include "arch-ppc64.h"
+#include "arch-riscv64.h"
 #include "arch-s390.h"
 #include "arch-s390x.h"
 
@@ -117,9 +119,11 @@ int main(int argc, char *argv[])
 			sys = mips64n32_syscall_iterate(iter);
 			break;
 		case SCMP_ARCH_PARISC:
-		case SCMP_ARCH_PARISC64:
 			sys = parisc_syscall_iterate(iter);
 			break;
+		case SCMP_ARCH_PARISC64:
+			sys = parisc64_syscall_iterate(iter);
+			break;
 		case SCMP_ARCH_PPC:
 			sys = ppc_syscall_iterate(iter);
 			break;
@@ -127,6 +131,9 @@ int main(int argc, char *argv[])
 		case SCMP_ARCH_PPC64LE:
 			sys = ppc64_syscall_iterate(iter);
 			break;
+		case SCMP_ARCH_RISCV64:
+			sys = riscv64_syscall_iterate(iter);
+			break;
 		case SCMP_ARCH_S390:
 			sys = s390_syscall_iterate(iter);
 			break;
